import numpy as np
from geospace.utils import ds_name


def grid_area(lon_res, lat_res, lat_upper, n_lat):
    """
    Calculate the area of a cell, in meters^2, on a lat/lon grid.

    This applies the following equation from Santini et al. 2010.

    S = (λ_2 - λ_1)(sinφ_2 - sinφ_1)R^2

    S = surface area of cell on sphere
    λ_1, λ_2 = authalic longitude in radians
    φ_1, φ_2 = authalic latitude in radians
    R = Earth's authalic radius

    Earth's authalic radius (meaning "equal area") is the
    radius of a hypothetical perfect sphere that has the
    same surface area as the reference ellipsoid.
    Authalic lon/lat are converted from geodetic lon/lat.

    Santini, M., Taramelli, A., & Sorichetta, A. (2010). ASPHAA:
    A GIS-Based Algorithm to Calculate Cell Area on a Latitude-
    Longitude (Geographic) Regular Grid. Transactions in GIS.
    https://doi.org/10.1111/j.1467-9671.2010.01200.x

    Args:
        lon_res float: resolution of longitude
        lat_res float: resolution of latitude
        lat_upper float: upper geodetic latitude
        n_lat int: number of a column grids that from the
        lat_upper to (lat_upper - lat_res * n_lat)

    Returns:
        np.ndarray[float]: area of the column grids
    """

    # semi-major axis and flattening of the WGS 84
    # https://en.wikipedia.org/wiki/World_Geodetic_System
    SMA = 6378137
    f = 1 / 298.257223563
    e = np.sqrt(2 * f - f * f)
    e1 = 1 - e * e
    e2 = 1 / (2 * e)
    ee = e * e
    sin90 = np.sin(np.pi / 2)
    q = e1 * (
        sin90 / (1 - ee * sin90 * sin90)
        - e2 * np.log((1 - e * sin90) / (1 + e * sin90))
    )
    R2 = SMA * SMA * q / 2

    lat_below = lat_upper - lat_res * n_lat
    sin_lats = np.sin(np.radians(np.linspace(lat_upper, lat_below, n_lat + 1)))
    q_lats = e1 * (
        sin_lats / (1 - ee * sin_lats * sin_lats)
        - e2 * np.log((1 - e * sin_lats) / (1 + e * sin_lats))
    )
    return R2 * np.radians(lon_res) * np.abs((q_lats[:-1] - q_lats[1:]) / q)


def area_per_row(trans, proj_wkt, n_rows, offset=0):
    if 'PROJCS' in proj_wkt:
        cell_area = abs(trans[1] * trans[5]) / 1000000
        return np.full(n_rows, cell_area)

    return (
        grid_area(abs(trans[1]), abs(trans[5]), trans[3] + offset * trans[5], n_rows)
        / 1000000
    )


def real_area(ds, rows, offset=None, return_row_area=False):
    rows = np.array(rows).flatten()
    if len(rows) == 0:
        return np.array([])
    ds = ds_name(ds)[0]
    trans = ds.GetGeoTransform()
    proj = ds.GetProjection()
    if 'PROJCS' in proj:
        cell_area = abs(trans[1] * trans[5]) / 1000000
        if return_row_area:
            return np.full(rows.shape, cell_area)
        return rows.shape[0] * cell_area

    if offset is None:
        offset = np.min(rows)
        rows = rows - offset
    row_counts = np.bincount(rows)
    row_area = (
        grid_area(
            abs(trans[1]),
            abs(trans[5]),
            trans[3] + offset * trans[5],
            row_counts.shape[0],
        )
        / 1000000
    )
    if return_row_area:
        return row_area[rows]
    return (row_counts * row_area).sum()


def distance(a, b):
    import math

    ELLIPSOIDS = {
        # model           major (km)   minor (km)     flattening
        'WGS-84': (6378.137, 6356.7523142, 1 / 298.257223563),
    }

    lng1, lat1 = math.radians(a[0]), math.radians(a[1])
    lng2, lat2 = math.radians(b[0]), math.radians(b[1])

    major, minor, f = ELLIPSOIDS['WGS-84']

    delta_lng = lng2 - lng1

    reduced_lat1 = math.atan((1 - f) * math.tan(lat1))
    reduced_lat2 = math.atan((1 - f) * math.tan(lat2))

    sin_reduced1, cos_reduced1 = math.sin(reduced_lat1), math.cos(reduced_lat1)
    sin_reduced2, cos_reduced2 = math.sin(reduced_lat2), math.cos(reduced_lat2)

    lambda_lng = delta_lng
    lambda_prime = 2 * math.pi

    iter_limit = 20

    i = 0

    while i == 0 or (abs(lambda_lng - lambda_prime) > 10e-12 and i <= iter_limit):
        i += 1

        sin_lambda_lng, cos_lambda_lng = math.sin(lambda_lng), math.cos(lambda_lng)

        sin_sigma = math.sqrt(
            (cos_reduced2 * sin_lambda_lng) ** 2
            + (
                cos_reduced1 * sin_reduced2
                - sin_reduced1 * cos_reduced2 * cos_lambda_lng
            )
            ** 2
        )

        if sin_sigma == 0:
            return 0  # Coincident points

        cos_sigma = (
            sin_reduced1 * sin_reduced2 + cos_reduced1 * cos_reduced2 * cos_lambda_lng
        )

        sigma = math.atan2(sin_sigma, cos_sigma)

        sin_alpha = cos_reduced1 * cos_reduced2 * sin_lambda_lng / sin_sigma
        cos_sq_alpha = 1 - sin_alpha**2

        if cos_sq_alpha != 0:
            cos2_sigma_m = cos_sigma - 2 * (sin_reduced1 * sin_reduced2 / cos_sq_alpha)
        else:
            cos2_sigma_m = 0.0  # Equatorial line

        C = f / 16.0 * cos_sq_alpha * (4 + f * (4 - 3 * cos_sq_alpha))

        lambda_prime = lambda_lng
        lambda_lng = delta_lng + (1 - C) * f * sin_alpha * (
            sigma
            + C
            * sin_sigma
            * (cos2_sigma_m + C * cos_sigma * (-1 + 2 * cos2_sigma_m**2))
        )

    if i > iter_limit:
        raise ValueError('Vincenty formula failed to converge!')

    u_sq = cos_sq_alpha * (major**2 - minor**2) / minor**2

    A = 1 + u_sq / 16384.0 * (4096 + u_sq * (-768 + u_sq * (320 - 175 * u_sq)))

    B = u_sq / 1024.0 * (256 + u_sq * (-128 + u_sq * (74 - 47 * u_sq)))

    delta_sigma = (
        B
        * sin_sigma
        * (
            cos2_sigma_m
            + B
            / 4.0
            * (
                cos_sigma * (-1 + 2 * cos2_sigma_m**2)
                - B
                / 6.0
                * cos2_sigma_m
                * (-3 + 4 * sin_sigma**2)
                * (-3 + 4 * cos2_sigma_m**2)
            )
        )
    )

    s = minor * A * (sigma - delta_sigma)
    return s


if __name__ == '__main__':
    area_columns = grid_area(1 / 3600, 1 / 3600, 60, 300000)
    area_columns = grid_area(0.5, 0.5, 60, 100)



import os
from osgeo import gdal


# set proj folder
_gdal_data = os.path.join(os.path.dirname(gdal.__file__), 'data')
os.environ.setdefault('PROJ_DATA', os.path.join(_gdal_data, 'proj'))
os.environ.setdefault('GDAL_DATA', os.path.join(_gdal_data, 'gdal'))
os.environ.setdefault('CPL_LOG', os.devnull)

# gdal config
gdal.DontUseExceptions()
gdal.SetConfigOption('SHAPE_ENCODING', 'utf-8')
gdal.PushErrorHandler('CPLQuietErrorHandler')
gdal.SetConfigOption('GDAL_FILENAME_IS_UTF8', 'YES')
gdal.SetConfigOption('GDAL_NUM_THREADS', 'ALL_CPUS')

# default spatial reference system
WGS84 = 'EPSG:4326'

# creation options
CREATION = [
    'BIGTIFF=YES',
    'TILED=YES',
    'NUM_THREADS=ALL_CPUS',
    'COMPRESS=ZSTD',
    'PREDICTOR=1',
    'ZSTD_LEVEL=1',
]

# mapping between gdal type and ogr field type
TYPE_MAP = {
    'bool': gdal.GDT_Byte,
    'uint8': gdal.GDT_Byte,
    'int8': gdal.GDT_Byte,
    'uint16': gdal.GDT_UInt16,
    'int16': gdal.GDT_Int16,
    'uint32': gdal.GDT_UInt32,
    'int32': gdal.GDT_Int32,
    'float32': gdal.GDT_Float32,
    'float64': gdal.GDT_Float64,
    'complex64': gdal.GDT_CFloat32,
    'complex128': gdal.GDT_CFloat64,
}

# cpu used
N_CPU = max(
    int(
        int(os.environ.get('SLURM_CPUS_PER_TASK', os.cpu_count()))
        * int(os.environ.get('SLURM_NTASKS', 1))
        * 3
        // 4
    ),
    1,
)



import numpy as np
from osgeo import gdal
from osgeo import gdal_array
from geospace.mask import shape_to_trans


def reproject(
    src_arr,
    dst_shape,
    src_transform,
    dst_transform,
    nodata=np.nan,
    **kwargs,
):
    srcSRS = kwargs.pop('srcSRS', 'EPSG:4326')
    dstSRS = kwargs.pop('dstSRS', 'EPSG:4326')
    resampleAlg = kwargs.pop('resampleAlg', gdal.GRA_Average)

    option = gdal.WarpOptions(
        srcNodata=nodata,
        dstNodata=nodata,
        srcSRS=srcSRS,
        dstSRS=dstSRS,
        multithread=True,
        resampleAlg=resampleAlg,
        **kwargs,
    )

    src_ds = gdal_array.OpenArray(src_arr)
    src_ds.SetGeoTransform(src_transform)

    dst_arr = np.full(dst_shape, nodata, dtype=src_arr.dtype)
    dst_ds = gdal_array.OpenArray(dst_arr)
    dst_ds.SetGeoTransform(dst_transform)

    gdal.Warp(dst_ds, src_ds, options=option)
    return dst_arr


def fill_nodata(ds, valid, nodata=np.nan, nearest=False, fast=True, **kwargs):
    if isinstance(ds, gdal.Dataset):
        trans, crs = ds.GetGeoTransform(), ds.GetProjection()
        arr = ds.ReadAsArray()
    else:
        trans, crs = shape_to_trans(*ds.shape[-2:]), 'EPSG:4326'
        arr = ds.copy()

    null = -9999 if np.isnan(nodata) else nodata
    arr[np.isnan(arr)] = null
    dst_ds = gdal_array.OpenArray(arr)
    dst_ds.SetGeoTransform(trans)
    dst_ds.SetProjection(crs)

    invalid = np.broadcast_to(~valid.astype(bool), arr.shape)
    mask = ((arr != null) | invalid) if fast else (arr != null)
    mask_ds = gdal_array.OpenArray(mask.astype(np.uint8))
    mask_ds.SetGeoTransform(trans)
    mask_ds.SetProjection(crs)

    kwargs.setdefault('maxSearchDist', 50)
    kwargs.setdefault('smoothingIterations', 0)
    kwargs.setdefault('options', {'NODATA': null, 'TEMP_FILE_DRIVER': 'MEM'})
    kwargs['options']['INTERPOLATION'] = 'NEAREST' if nearest else 'INV_DIST'

    for i in range(dst_ds.RasterCount):
        band = dst_ds.GetRasterBand(i + 1)
        band.SetNoDataValue(null)
        maskBand = mask_ds.GetRasterBand(i + 1)
        gdal.FillNodata(targetBand=band, maskBand=maskBand, **kwargs)

    if (arr[~(mask | invalid)] == null).any():
        if not fast:
            # Some nodata not filled, consider larger maxSearchDist
            kwargs['maxSearchDist'] *= 2
        arr = fill_nodata(ds, valid, nodata, nearest, fast=False, **kwargs)
    arr[invalid] = nodata
    return arr


if __name__ == '__main__':
    # Example usage
    arr = reproject(
        np.zeros((3600, 7200), dtype=np.float32),
        (1800, 3600),
        (-180, 0.05, 0, 90, 0, -0.05),
        (-180, 0.1, 0, 90, 0, -0.1),
    )



import numpy as np
from osgeo import ogr
from geospace._const import WGS84
from geospace.utils import geo2imagexy, ds_name
from geospace.projection import read_srs, coord_trans


def _prj_bound(ds, bound, bound_srs):
    x_min, y_min, x_max, y_max = bound
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(min(x_min, x_max), min(y_min, y_max))
    ring.AddPoint(max(x_min, x_max), min(y_min, y_max))
    ring.AddPoint(max(x_min, x_max), max(y_min, y_max))
    ring.AddPoint(min(x_min, x_max), max(y_min, y_max))
    ring.AddPoint(min(x_min, x_max), min(y_min, y_max))

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)

    trans = coord_trans(bound_srs, ds)
    trans_reverse = coord_trans(ds, bound_srs)

    # create a geometry from coordinates
    t = ds.GetGeoTransform()
    point = ogr.Geometry(ogr.wkbPoint)
    point_dx = ogr.Geometry(ogr.wkbPoint)
    point_dy = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(
        t[0] + ds.RasterXSize // 2 * t[1] + ds.RasterYSize // 2 * t[2],
        t[3] + ds.RasterXSize // 2 * t[4] + ds.RasterYSize // 2 * t[5],
    )
    point_dx.AddPoint(
        t[0] + (ds.RasterXSize // 2 + 1) * t[1] + ds.RasterYSize // 2 * t[2],
        t[3] + (ds.RasterXSize // 2 + 1) * t[4] + ds.RasterYSize // 2 * t[5],
    )
    point_dy.AddPoint(
        t[0] + ds.RasterXSize // 2 * t[1] + (ds.RasterYSize // 2 + 1) * t[2],
        t[3] + ds.RasterXSize // 2 * t[4] + (ds.RasterYSize // 2 + 1) * t[5],
    )
    point.Transform(trans_reverse)
    point_dx.Transform(trans_reverse)
    point_dy.Transform(trans_reverse)
    dx = abs(point.GetPoint()[0] - point_dx.GetPoint()[0])
    dy = abs(point.GetPoint()[1] - point_dy.GetPoint()[1])

    # density geom
    geom = ogr.CreateGeometryFromWkb(poly.ExportToWkb())
    geom.Segmentize(min(dx, dy) / 2)
    geom.Transform(trans)

    # get boundary
    win = np.array([geom.GetEnvelope()])
    bound = [win[:, 0:2].min(), win[:, 2:].min(), win[:, 0:2].max(), win[:, 2:].max()]
    return bound


def _enlarge_bound(ds, x_min, y_min, x_max, y_max):
    t = ds.GetGeoTransform()
    ulX, ulY = geo2imagexy(ds, x_min, y_min)
    lrX, lrY = geo2imagexy(ds, x_max, y_max)
    clip_range = [
        np.min([ulX, lrX], axis=0),
        np.min([ulY, lrY], axis=0),
        (np.abs(ulX - lrX) + 1),
        (np.abs(ulY - lrY) + 1),
    ]
    ul_lon = t[0] + t[1] * clip_range[0] + t[2] * clip_range[1]
    ul_lat = t[3] + t[4] * clip_range[0] + t[5] * clip_range[1]
    lr_lon = (
        t[0]
        + t[1] * (clip_range[0] + clip_range[2])
        + t[2] * (clip_range[1] + clip_range[3])
    )
    lr_lat = (
        t[3]
        + t[4] * (clip_range[0] + clip_range[2])
        + t[5] * (clip_range[1] + clip_range[3])
    )
    bound = [
        np.min([ul_lon, lr_lon], axis=0),
        np.min([ul_lat, lr_lat], axis=0),
        np.max([ul_lon, lr_lon], axis=0),
        np.max([ul_lat, lr_lat], axis=0),
    ]

    return bound, clip_range


def bound_raster(ds, bound, bound_srs=WGS84):
    ds, _ = ds_name(ds)
    x_min, y_min, x_max, y_max = _prj_bound(ds, bound, bound_srs)
    bound = _enlarge_bound(ds, x_min, y_min, x_max, y_max)[0]

    return bound, read_srs(ds).ExportToProj4()


def bound_layers(layers):
    # clip extent
    if not isinstance(layers, list):
        layers = [layers]
    if isinstance(layers[0], str):
        shp_in = [ogr.Open(lyr) for lyr in layers]
        layers = [s.GetLayer() for s in shp_in]
    win = np.array([lyr.GetExtent() for lyr in layers])
    bound = [win[:, 0:2].min(), win[:, 2:].min(), win[:, 0:2].max(), win[:, 2:].max()]

    return bound, read_srs(layers).ExportToProj4()


def grid_bound(ds, regions):
    """return gridded bound whose resolution exactly matches that of ds

    Args:
        ds (gdal.dataset): dataset or path
        regions (list or str): regions of interest

    Returns:
        bound: spatial range of the regions in the ds
        crs_transform: like bound but used in GEE
    """
    ds = ds_name(ds)[0]
    t = ds.GetGeoTransform()
    bound, bound_srs = bound_layers(regions)
    bound = bound_raster(ds, bound, bound_srs)[0]
    crs_transform = [t[1], 0, bound[0], 0, t[5], bound[-1]]
    return bound, crs_transform



import os
import numpy as np
from pathlib import Path
from osgeo import gdal, ogr
from geospace.projection import read_srs
from geospace.shape import shp_projection
from geospace._const import WGS84, CREATION
from geospace.utils import imagexy2geo, ds_name, context_file, zeros_tif


def shp2ras(
    shp,
    out_path,
    trans,
    attr='logK_Ice_x',
    nodata=0,
    rasterize_option=['ALL_TOUCHED=TRUE'],
):
    out_file = context_file(shp, out_path)

    if Path(out_file).exists():
        return

    # create tif
    zeros_tif(
        out_file,
        round(abs(2 * trans[0] / trans[1])),
        round(abs(2 * trans[3] / trans[5])),
        1,
        gdal.GDT_Float64,
        trans,
        'EPSG:4326',
        nodata=nodata,
    )
    ds = gdal.Open(out_file, gdal.GA_Update)

    # rasterize
    ds_shp = ogr.Open(shp)
    layer = ds_shp.GetLayer()
    gdal.RasterizeLayer(
        ds, [1], layer, options=['ATTRIBUTE=%s' % attr] + rasterize_option
    )

    # deal with permeability units
    if Path(out_file).stem == 'permeability':
        arr = ds.ReadAsArray()
        arr[arr != nodata] = arr[arr != nodata] / 100
        ds.WriteArray(arr)
    ds = None


def rasterize(
    shp,
    attr,
    out_path,
    ds_eg,
    tem_path,
    rasterize_option=['ALL_TOUCHED=TRUE'],
    **kwargs,
):
    # create out put name
    out_file = os.path.join(
        out_path, os.path.splitext(os.path.basename(shp))[0] + '.tif'
    )
    if os.path.exists(out_file):
        return
    tem_file = os.path.join(
        tem_path, os.path.splitext(os.path.basename(shp))[0] + '.tif'
    )

    # extent warp options
    ds_ex = gdal.Translate('/vsimem/_extent.tif', ds_eg, bandList=[1])
    t = ds_eg.GetGeoTransform()
    temp_option = gdal.WarpOptions(
        multithread=True,
        creationOptions=CREATION,
        xRes=t[1] / 10,
        yRes=t[5] / 10,
        outputType=gdal.GDT_Float64,
        **kwargs,
    )

    ds_tem = gdal.Warp(tem_file, ds_ex, options=temp_option)
    band = ds_tem.GetRasterBand(1)
    option = gdal.WarpOptions(
        multithread=True,
        creationOptions=CREATION,
        xRes=t[1],
        yRes=t[5],
        resampleAlg=gdal.GRA_Average,
        outputType=gdal.GDT_Float64,
        **kwargs,
    )

    shp_factor = ogr.Open(shp)
    layer = shp_factor.GetLayer()

    # create and use RasterizeLayer
    band.Fill(band.GetNoDataValue())
    gdal.RasterizeLayer(
        ds_tem, [1], layer, options=['ATTRIBUTE=%s' % attr] + rasterize_option
    )
    gdal.Warp(out_file, ds_tem, options=option)

    return out_file


def download_tiles(shp, tile_pixel):
    # create the output layer
    out_shp = '/vsimem/outline_wgs84.shp'
    shp_projection(shp, out_shp)
    outDataSet = ogr.Open(out_shp)
    outLayer = outDataSet.GetLayer()

    # Create the destination data source
    x_res = int(360 / tile_pixel)
    y_res = int(180 / tile_pixel)
    target_ds = gdal.GetDriverByName('GTiff').Create(
        '/vsimem/_tile.tif', x_res, y_res, 1, gdal.GDT_Byte
    )
    target_ds.SetGeoTransform((-180, tile_pixel, 0, 90, 0, -tile_pixel))
    target_ds.SetSpatialRef(read_srs(WGS84))
    band = target_ds.GetRasterBand(1)
    band.SetNoDataValue(0)

    # Rasterize
    gdal.RasterizeLayer(
        target_ds, [1], outLayer, burn_values=[1], options=['ALL_TOUCHED=TRUE']
    )
    burn_tiles = target_ds.GetRasterBand(1).ReadAsArray()
    rows, cols = np.where(burn_tiles == 1)
    if rows.shape == 0:
        return None
    lon, lat = np.int_(imagexy2geo(target_ds, rows - 0.5, cols - 0.5))
    return lon, lat


def masked_outside(shp, ds):
    import psutil

    ds, ras = ds_name(ds)
    ds = gdal.Open(ras, gdal.GA_Update)
    trans = ds.GetGeoTransform()
    ratio = int(
        ds.RasterYSize
        * ds.RasterXSize
        / (
            psutil.virtual_memory().available
            * 0.5
            / (2 + ds.ReadAsArray(0, 0, 1, 1).dtype.itemsize)
        )
        + 1
    )

    # create the output layer
    out_shp = '/vsimem/outline_wgs84.shp'
    shp_projection(shp, out_shp, out_srs=read_srs(ds))
    outDataSet = ogr.Open(out_shp)
    outLayer = outDataSet.GetLayer()

    band = ds.GetRasterBand(1)
    b_xsize, b_ysize = band.GetBlockSize()
    xsize = max(ds.RasterXSize // ratio + 1, b_xsize)
    ysize = max(ds.RasterYSize // ratio + 1, b_ysize)
    xoffs = range(0, ds.RasterXSize, xsize)
    yoffs = range(0, ds.RasterYSize, ysize)
    for i in range(ds.RasterCount):
        band = ds.GetRasterBand(i + 1)
        for xoff in xoffs:
            for yoff in yoffs:
                win_xsize = min(xsize, ds.RasterXSize - xoff)
                win_ysize = min(ysize, ds.RasterYSize - yoff)

                # Create the destination data source
                target_ds = gdal.GetDriverByName('GTiff').Create(
                    '/vsimem/_outside.tif', win_xsize, win_ysize, 1, gdal.GDT_Byte
                )
                left_up_lon, left_up_lat = imagexy2geo(ds, yoff - 0.5, xoff - 0.5)
                target_ds.SetGeoTransform(
                    (left_up_lon, trans[1], 0, left_up_lat, 0, trans[5])
                )
                target_ds.SetProjection(ds.GetProjection())
                target_band = target_ds.GetRasterBand(1)
                target_band.SetNoDataValue(0)

                # Rasterize
                gdal.RasterizeLayer(
                    target_ds,
                    [1],
                    outLayer,
                    burn_values=[1],
                    options=['ALL_TOUCHED=TRUE'],
                )
                outside = np.logical_not(target_ds.GetRasterBand(1).ReadAsArray())
                target_band = None
                target_ds = None

                if outside.any():
                    arr = ds.ReadAsArray(
                        xoff=xoff, yoff=yoff, xsize=win_xsize, ysize=win_ysize
                    )
                    arr[outside] = band.GetNoDataValue()
                    band.WriteArray(arr, xoff=xoff, yoff=yoff)
                    del arr
                del outside
        band = None



from ._const import N_CPU, WGS84, CREATION, TYPE_MAP
from .array import reproject, fill_nodata
from .boundary import bound_raster, bound_layers, grid_bound
from .calculator import map_calc
from .mask import shape_to_trans, land_mask
from .projection import read_srs, coord_trans
from .ras_to_shp import polygonize
from .shp_to_ras import shp2ras, rasterize, download_tiles, masked_outside
from .spatial_calc import grid_area, area_per_row, real_area, distance
from .zonal_stats import extract, basin_average

from .raster import (
    convert_uint8,
    resample,
    mosaic,
    project_raster,
    grib_to_tif,
    tif_copy_assign,
)
from .shape import (
    shp_buffer,
    shp_projection,
    shp_filter,
    shp_geom_map,
    shp_weighted_mean,
)
from .utils import (
    block_write,
    rep_file,
    rep_name,
    geo2imagexy,
    imagexy2geo,
    meshgrid,
    context_file,
    ds_name,
    get_extent,
    zeros_tif,
)
from .gee_export import (
    gee_init,
    gee_export_tif,
    gee_export_csv,
    gee_to_drive,
    gee_soilgrids,
    gee_wind,
    gee_group_by_month,
    gee_seasonality_index,
)



import os
import numpy as np
from osgeo import gdal
from pathlib import Path
from geospace.projection import read_srs
from geospace._const import WGS84, CREATION, TYPE_MAP
from geospace.utils import rep_file, ds_name, context_file


def convert_uint8(ds, in_nodata=None, out_nodata=255):
    ds, ras = ds_name(ds)
    frist_band = ds.GetRasterBand(1)

    if (
        frist_band.DataType != gdal.GDT_Byte
        or frist_band.ReadAsArray(0, 0, 1, 1).dtype != np.int8
    ):
        return ras

    if frist_band.GetNoDataValue() is not None:
        in_nodata = frist_band.GetNoDataValue()
    if in_nodata is None:
        raise (ValueError('in_nodata must be initialed'))

    option = gdal.WarpOptions(
        multithread=True,
        creationOptions=CREATION,
        srcNodata=in_nodata,
        dstNodata=out_nodata,
        outputType=gdal.GDT_Byte,
    )
    out_file = rep_file(os.path.dirname(ras), ras)
    gdal.Warp(out_file, ras, options=option)

    ds = None
    gdal.GetDriverByName('GTiff').Delete(ras)
    os.rename(out_file, ras)

    return ras


def resample(ds, out_path, **kwargs):
    ds, ras = ds_name(ds)
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    resample_alg = kwargs.pop('resampleAlg', gdal.GRA_Average)
    option = gdal.WarpOptions(
        multithread=True, creationOptions=CREATION, resampleAlg=resample_alg, **kwargs
    )
    gdal.Warp(out_file, ds, options=option)

    return out_file


def mosaic(rasters, out_path, **kwargs):
    if isinstance(rasters, (str, Path)):
        rasters = [str(rasters)]
    else:
        rasters = [str(ras) for ras in rasters]

    ds, ras = ds_name(rasters[0])
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    # convert longitude [0, 360] to [-180, 180] when mosaic nc files
    all_nc = all(Path(f).suffix == '.nc' for f in rasters)
    separate = kwargs.pop('separate', True if all_nc else False)
    srcSRS = read_srs([ds, kwargs.pop('srcSRS', WGS84)])
    resample_alg = kwargs.pop('resampleAlg', gdal.GRA_Average)
    outputBounds = kwargs.pop('outputBounds', [-180, -90, 180, 90] if all_nc else None)
    option = gdal.WarpOptions(
        multithread=True,
        srcSRS=srcSRS,
        creationOptions=CREATION,
        resampleAlg=resample_alg,
        outputBounds=outputBounds,
        **kwargs,
    )

    ds = gdal.BuildVRT('/vsimem/Mosaic.vrt', rasters, separate=separate)
    ds_out = gdal.Warp(out_file, ds, options=option)

    if separate:
        # each raster only have one band in the mosaic
        band_names = (Path(p).stem for p in rasters)
        [
            ds_out.GetRasterBand(i + 1).SetDescription(band_name)
            for i, band_name in enumerate(band_names)
        ]

    return out_file


def project_raster(ds, out_path, **kwargs):
    ds, ras = ds_name(ds)
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    # input SpatialReference
    in_srs = kwargs.pop('srcSRS', WGS84)
    inSpatialRef = read_srs([ds, in_srs])

    # output SpatialReference
    out_srs = kwargs.pop('dstSRS', WGS84)
    outSpatialRef = read_srs(out_srs)

    resample_alg = kwargs.pop('resampleAlg', gdal.GRA_Average)
    option = gdal.WarpOptions(
        creationOptions=CREATION,
        resampleAlg=resample_alg,
        srcSRS=inSpatialRef,
        dstSRS=outSpatialRef,
        multithread=True,
        **kwargs,
    )
    gdal.Warp(out_file, ds, options=option)

    return out_file


def grib_to_tif(ds, out_path=None, **kwargs):
    ds, ras = ds_name(ds)

    if os.path.splitext(os.path.basename(ras))[1] != '.grib':
        return

    if out_path:
        out_file = context_file(ras, out_path)
    else:
        out_file = os.path.join(
            os.path.dirname(ras), os.path.splitext(os.path.basename(ras))[0] + '.tif'
        )

    if os.path.exists(out_file):
        return out_file

    srcSRS = read_srs([ds, kwargs.pop('srcSRS', WGS84)])
    option = gdal.WarpOptions(
        multithread=True, srcSRS=srcSRS, creationOptions=CREATION, **kwargs
    )
    gdal.Warp(out_file, ds, options=option)

    return out_file


def tif_copy_assign(out_file, ds_eg, array, srs=None, nodata=None):
    out_file = str(out_file)
    if os.path.exists(out_file):
        return out_file
    ds_eg = ds_name(ds_eg)[0]

    if array.ndim == 2:
        array = array.reshape([1, *array.shape])
    if array.ndim != 3:
        raise (Exception('array dims must be 2 or 3'))

    # set nodata value
    if nodata is None:
        if ds_eg.GetRasterBand(1).GetNoDataValue() is not None:
            nodata = ds_eg.GetRasterBand(1).GetNoDataValue()
        else:
            raise (Exception('nodata must be passed'))
    if isinstance(array, np.ma.MaskedArray):
        array.set_fill_value(nodata)
        array = array.filled()

    ds = gdal.GetDriverByName('GTiff').Create(
        out_file,
        array.shape[2],
        array.shape[1],
        array.shape[0],
        TYPE_MAP[array.dtype.name],
        CREATION,
    )

    # fill with array
    band = ds.GetRasterBand(1)
    band.SetNoDataValue(nodata)
    ds.WriteArray(array)

    # set geotransform
    trans = ds_eg.GetGeoTransform()
    ds.SetGeoTransform(tuple(trans))

    # set SpatialReference
    ds.SetSpatialRef(read_srs([srs, ds_eg]))

    return out_file



from typing import List
from pathlib import Path
from osgeo import gdal, ogr, osr


def read_srs(srs):
    if isinstance(srs, osr.SpatialReference):
        return srs
    elif isinstance(srs, str):
        if Path(srs).is_file() or 'vsimem' in str(Path(srs).parent):
            ds = gdal.Open(srs)
            if ds is not None:
                return osr.SpatialReference(wkt=ds.GetProjection())
            ds = ogr.Open(srs)
            if ds is not None:
                layer = ds.GetLayer()
                return layer.GetSpatialRef()
        else:
            SpatialRef = osr.SpatialReference(wkt=None)
            if SpatialRef.SetFromUserInput(srs) == 0:
                return SpatialRef
    elif isinstance(srs, gdal.Dataset) or isinstance(srs, ogr.DataSource):
        return read_srs(srs.GetDescription())
    elif isinstance(srs, ogr.Layer):
        return srs.GetSpatialRef()
    elif isinstance(srs, List):
        for s in srs:
            SpatialRef = read_srs(s)
            if SpatialRef is not None and SpatialRef.ExportToWkt() != '':
                return SpatialRef
        raise (ValueError('srs must be set'))
    return None


def coord_trans(in_srs, out_srs):
    inSpatialRef, outSpatialRef = read_srs(in_srs), read_srs(out_srs)
    # create the CoordinateTransformation
    if int(gdal.__version__[0]) >= 3:
        outSpatialRef.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    coordTrans = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)
    return coordTrans



import numpy as np
from osgeo import gdal, ogr
from geospace.utils import zeros_tif
from geospace.shape import shp_filter


def rounder(x, n=2):
    """Round a value to n significant figures.

    If x is an integer or have only zeros after the decimal point, no need to round.
    If x is a float, keep only the first n digits other than zero after the decimal point
    """
    import math

    try:
        return round(x, -int(math.floor(math.log10(abs(math.modf(x)[0])))) + n - 1)
    except ValueError:
        return x


def shape_to_trans(y_size, x_size):
    # sizes=3600x1801 special case for ERA5
    res = rounder(360 / x_size)
    res = 360 / x_size if res < 0.01 else res
    trans = (
        -rounder(res * x_size / 2, 3),
        res,
        0.0,
        rounder(res * y_size / 2, 3),
        0.0,
        -res,
    )
    return trans


def land_mask(
    out_file='/vsimem/land.tif',
    sizes='3600x1800',
    greenland=[126],
    exclude_glacier=True,
):
    import re
    import math
    import cartopy
    from skimage.morphology import remove_small_holes, remove_small_objects

    if gdal.Open(out_file) is not None:
        return out_file

    x_size, y_size = (int(size) for size in re.findall(r'\d+', sizes))
    trans = shape_to_trans(y_size, x_size)
    n_band, data_type, srs, res = 1, gdal.GDT_Byte, 'EPSG:4326', trans[1]

    # mask for lake
    f = '/vsimem/_lake.tif'
    shp = shp_filter(
        cartopy.io.shapereader.natural_earth(name='lakes'), filter_sql="scalerank = '0'"
    )
    ds_shp = ogr.Open(shp)
    layer = ds_shp.GetLayer()
    zeros_tif(f, x_size, y_size, n_band, data_type, trans, srs)
    ds = gdal.Open(f, gdal.GA_Update)
    gdal.RasterizeLayer(ds, [1], layer, burn_values=[1])
    ds = None
    lake = gdal.Open(f).ReadAsArray().astype(bool)

    # mask for Antarctica and Greenland
    f = '/vsimem/_glacier.tif'
    if exclude_glacier:
        sql = "','".join(str(i) for i in list(range(0, 8)) + greenland)
        shp = shp_filter(
            cartopy.io.shapereader.natural_earth(name='land'),
            filter_sql=f"FID IN ('{sql}')",
        )
        ds_shp = ogr.Open(shp)
        layer = ds_shp.GetLayer()
        zeros_tif(f, x_size, y_size, n_band, data_type, trans, srs)
        ds = gdal.Open(f, gdal.GA_Update)
        gdal.RasterizeLayer(
            ds, [1], layer, burn_values=[1], options=['ALL_TOUCHED=TRUE']
        )
        ds = None
        glacier = gdal.Open(f).ReadAsArray().astype(bool)
    else:
        glacier = np.full(lake.shape, False)

    # mask for natureal earth land
    ds_shp = ogr.Open(cartopy.io.shapereader.natural_earth(name='land'))
    layer = ds_shp.GetLayer()
    zeros_tif(out_file, x_size, y_size, n_band, data_type, trans, srs)
    ds = gdal.Open(out_file, gdal.GA_Update)
    gdal.RasterizeLayer(ds, [1], layer, burn_values=[1], options=['ALL_TOUCHED=TRUE'])
    ds = None
    land = gdal.Open(out_file).ReadAsArray().astype(bool)

    # final mask
    ds = gdal.Open(out_file, gdal.GA_Update)
    valid = ~lake & ~glacier & land
    # valid = erosion(dilation(erosion(valid, disk(2)), disk(4)), disk(2))
    # fill or remove 30 grids for 0.1
    thres = math.ceil(0.3 / (res**2))

    # morphology
    valid = remove_small_objects(valid, thres)
    valid = remove_small_holes(valid, thres)

    # save to tif
    ds.WriteArray(valid)
    ds = None
    return out_file



import os
import numpy as np
from osgeo import ogr
from geospace._const import WGS84
from geospace.utils import rep_file
from geospace.projection import read_srs, coord_trans


def shp_buffer(in_shp, out_shp, buffdist, in_srs=None):
    return shp_geom_map(
        in_shp, out_shp, in_srs=in_srs, func=lambda geom: geom.Buffer(float(buffdist))
    )


def shp_projection(in_shp, out_shp, in_srs=WGS84, out_srs=WGS84):
    # Filename of input OGR file
    if isinstance(in_shp, ogr.Layer):
        inLayer = in_shp
    else:
        source_ds = ogr.Open(in_shp)
        inLayer = source_ds.GetLayer()

    # input SpatialReference
    inSpatialRef = read_srs([inLayer, in_srs])

    # output SpatialReference
    outSpatialRef = read_srs(out_srs)

    # create the CoordinateTransformation
    coordTrans = coord_trans(inSpatialRef, outSpatialRef)

    def transform(geom):
        geom.Transform(coordTrans)
        return geom

    return shp_geom_map(in_shp, out_shp, out_srs=outSpatialRef, func=transform)


def shp_filter(shps, filter_sql, filter_shp=None):
    ds_shp = ogr.Open(shps, 0)
    layer = ds_shp.GetLayer()
    layer.SetAttributeFilter(filter_sql)

    if filter_shp is False:
        driver = ogr.GetDriverByName('MEM')
        out_ds = driver.CreateDataSource('')
        out_ds.CopyLayer(layer, 'filter')
        return out_ds

    driver = ogr.GetDriverByName('ESRI Shapefile')
    if filter_shp is None:
        out_shp = '/vsimem/filter.shp'
    else:
        out_shp = filter_shp

    out_ds = driver.CreateDataSource(out_shp)
    out_ds.CopyLayer(layer, 'filter')

    return out_shp


def shp_geom_map(in_shp, out_shp, idxs=None, func=None, in_srs=None, out_srs=None):
    # Filename of input OGR file
    if isinstance(in_shp, ogr.Layer):
        inLayer = in_shp
    else:
        source_ds = ogr.Open(in_shp)
        inLayer = source_ds.GetLayer()

    # output SpatialReference
    outSpatialRef = read_srs([out_srs, inLayer, in_srs])

    # create the output layer
    return_ds = out_shp is None or out_shp is False
    if return_ds:
        driver = ogr.GetDriverByName('MEM')
        outDataSet = driver.CreateDataSource('')
        outLayer = outDataSet.CreateLayer('memory', outSpatialRef)
    else:
        driver = ogr.GetDriverByName('ESRI Shapefile')
        if 'vsimem' not in os.path.dirname(out_shp):
            if (
                not os.path.exists(os.path.dirname(out_shp))
                and os.path.dirname(out_shp) != ''
            ):
                os.makedirs(os.path.dirname(out_shp))
        if os.path.exists(out_shp):
            driver.DeleteDataSource(out_shp)
        outDataSet = driver.CreateDataSource(out_shp)
        outLayer = outDataSet.CreateLayer(out_shp, outSpatialRef)

    # add fields
    inLayerDefn = inLayer.GetLayerDefn()
    for i in range(0, inLayerDefn.GetFieldCount()):
        fieldDefn = inLayerDefn.GetFieldDefn(i)
        outLayer.CreateField(fieldDefn)

    # get the output layer's feature definition
    outLayerDefn = outLayer.GetLayerDefn()

    # set all feature idxs if not assigned
    if idxs is None:
        idxs = range(inLayer.GetFeatureCount())
    elif isinstance(idxs, int):
        idxs = [idxs]

    # loop through the input features
    for i in idxs:
        inFeature = inLayer.GetFeature(i)
        # get the input geometry
        geom = inFeature.GetGeometryRef()
        # create a new feature
        outFeature = ogr.Feature(outLayerDefn)
        # set the geometry and attribute
        outFeature.SetGeometry(geom if func is None else func(geom))
        for j in range(0, outLayerDefn.GetFieldCount()):
            outFeature.SetField(
                outLayerDefn.GetFieldDefn(j).GetNameRef(), inFeature.GetField(j)
            )
        # add the feature to the shapefile
        outLayer.CreateFeature(outFeature)
        # dereference the features and get the next input feature
        outFeature = None

    if return_ds:
        return outDataSet
    else:
        return out_shp


def shp_weighted_mean(in_shp, clip_shp, field, out_shp=None, save_cache=False):
    driver = ogr.GetDriverByName('ESRI Shapefile')

    # get layer of in_shp
    if isinstance(in_shp, str):
        ds = ogr.Open(in_shp)
    else:
        ds = in_shp

    in_layer = ds.GetLayer()
    srs = in_layer.GetSpatialRef()

    # project clip_shp
    if save_cache:
        proj_shp = rep_file(
            'cache', os.path.splitext(os.path.basename(clip_shp))[0] + '_proj.shp'
        )
    else:
        proj_shp = '/vsimem/_proj.shp'
    shp_projection(clip_shp, proj_shp, out_srs=srs)
    clip_ds = ogr.Open(proj_shp)
    clip_layer = clip_ds.GetLayer()

    # export out_shp
    if out_shp is None:
        if save_cache:
            out_shp = rep_file(
                'cache', os.path.splitext(os.path.basename(clip_shp))[0] + '_out.shp'
            )
        else:
            out_shp = '/vsimem/_out.shp'

    out_ds = driver.CreateDataSource(out_shp)
    out_layer = out_ds.CreateLayer(out_shp, srs=srs)

    in_layer.Clip(clip_layer, out_layer)

    area = []
    logK = []
    # newField = ogr.FieldDefn('Area', ogr.OFTReal)
    # out_layer.CreateField(newField)
    c = out_layer.GetFeatureCount()
    for i in range(c):
        f = out_layer.GetFeature(i)
        area.append(f.GetGeometryRef().GetArea())
        logK.append(f.GetField(field))
        # f.SetField('Area', f.GetGeometryRef().GetArea())
        # out_layer.SetFeature(f)
    area = np.array(area)
    logK = np.array(logK)
    mean_logK = np.average(logK, weights=area)
    # mean_logK = np.log10(np.average(np.power(10, logK / 100), weights=area))
    out_layer = None
    out_ds = None
    return mean_logK



import os


def connected_to_internet(url='http://www.google.com/', timeout=1):
    import requests

    try:
        _ = requests.head(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        print('Failed to access GEE')
    return False


# gee initialization
def gee_init(proxy='http://127.0.0.1:7890'):
    import ee

    if not connected_to_internet():
        print('Set proxy to 127.0.0.1:7890')
        os.environ['HTTP_PROXY'] = proxy
        os.environ['HTTPS_PROXY'] = proxy
    try:
        ee.Initialize()
    except Exception:
        ee.Authenticate()
        ee.Initialize()


def gee_export_tif(image, filename, timeout=300, **params):
    """Export as tif, must be the original image, instead of the reprojected one

    Args:
        fc (ee.FeatureCollection): the spatial scope of the exported tif
        image (ee.Image): the image that would be exported
        filename (string): exported path
        crs (str, optional): A default CRS string to use for any bands that do not explicitly specify one
        crs_transform ([type]): control the spatial resolution and alignment, e.g., [1, 0, -180, 0, -1, 90]
        region (object, optional): A polygon specifying a region to download
        filePerBand (bool, optional): Whether to produce a different GeoTIFF per band

        details are in ee.Image.getDownloadURL()
    """
    import ee
    import requests
    import zipfile

    if not isinstance(image, ee.Image):
        print('The image must be an ee.Image.')
        return

    filename = os.path.abspath(filename)
    basename = os.path.basename(filename)
    name = os.path.splitext(basename)[0]
    filetype = os.path.splitext(basename)[1][1:].lower()
    filename_zip = filename.replace('.tif', '.zip')

    if filetype != 'tif':
        print('The filename must end with .tif')
        return

    try:
        print('Generating URL ...')
        params['name'] = name
        params['filePerBand'] = params.pop('filePerBand', False)
        if params.get('region') == 'global':
            params['region'] = image.geometry()

        try:
            url = image.getDownloadURL(params)
        except Exception as e:
            print('An error occurred while downloading.')
            print(e)
            return
        print(f'Downloading data from {url}\nPlease wait ...')
        r = requests.get(url, stream=True, timeout=timeout)

        if r.status_code != 200:
            print('An error occurred while downloading.')
            return

        with open(filename_zip, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=1024):
                fd.write(chunk)

    except Exception as e:
        print('An error occurred while downloading.')
        print(r.json()['error']['message'])
        return

    try:
        with zipfile.ZipFile(filename_zip) as z:
            z.extractall(os.path.dirname(filename))
        os.remove(filename_zip)

        if params['filePerBand']:
            print(f'Data downloaded to {os.path.dirname(filename)}')
        else:
            print(f'Data downloaded to {filename}')
    except Exception as e:
        print(e)


def gee_to_drive(image, **params):
    """wrapper for GEE Export.image.toDrive

    Args:
        image: The image to be exported.
        description: Human-readable name of the task.
        folder: The name of a unique folder in your Drive account to
            export into. Defaults to the root of the drive.
        dimensions: The dimensions of the exported image. Takes either a
            single positive integer as the maximum dimension or "WIDTHxHEIGHT"
            where WIDTH and HEIGHT are each positive integers.
        region: The lon,lat coordinates for a LinearRing or Polygon
            specifying the region to export. Can be specified as a nested
            lists of numbers or a serialized string. Defaults to the image's
            region.
        crs: The coordinate reference system of the exported image's
            projection. Defaults to the image's default projection.
        crsTransform: A comma-separated string of 6 numbers describing
            the affine transform of the coordinate reference system of the
            exported image's projection, in the order: xScale, xShearing,
            xTranslation, yShearing, yScale and yTranslation. Defaults to
            the image's native CRS transform.
        maxPixels: The maximum allowed number of pixels in the exported
            image. The task will fail if the exported region covers more
            pixels in the specified projection. Defaults to 100,000,000.

        details are in Export.image.toDrive()
    """
    import ee

    params['image'] = image
    params['description'] = params.pop(
        'description', image.bandNames().get(0).getInfo()
    )
    params['crs'] = params.pop('crs', 'EPSG:4326')
    params['maxPixels'] = params.pop('maxPixels', 1e13)

    if params.get('crsTransform') == 'ERA5_LAND':
        params['crsTransform'] = [0.1, 0, -180.05, 0, -0.1, 90.05]

    if params.get('region') == 'global':
        params['region'] = ee.Geometry.Rectangle(
            [-180, -90, 180, 90], geodesic=False, proj='EPSG:4326'
        )
        if params.get('dimensions') == 'default':
            params['dimensions'] = '360x180'
    elif params.get('region') == 'nonATA':
        params['region'] = ee.Geometry.Rectangle(
            [-180, -60, 180, 89], geodesic=False, proj='EPSG:4326'
        )
        if params.get('dimensions') == 'default':
            params['dimensions'] = '360x149'

    ee.batch.Export.image.toDrive(**params).start()


def gee_export_csv(
    fc, image, fields=['STAID', '.*mean'], return_url=False, to_drive=None, **kwargs
):
    """export a csv containing the basin average value

    Args:
        fc (ee.FeatureCollection): e.g. basins
        image (ee.Image): e.g. DEM

    Returns:
        DataFrame: it has fields of 'STAID' and 'mean'
    """
    import ee
    import pandas as pd

    # export as csv
    reducer = kwargs.pop('reducer', ee.Reducer.mean())
    scale = kwargs.pop('scale', image.projection().nominalScale())

    if image.projection().crs().getInfo() != 'EPSG:4326':
        image = image.reproject('EPSG:4326', None, scale)

    means = image.reduceRegions(fc, reducer=reducer, scale=scale, **kwargs)

    if to_drive is None:
        url = means.select(fields, retainGeometry=False).getDownloadURL(filetype='csv')
        if return_url:
            return url
        else:
            return pd.read_csv(url)
    else:
        ee.batch.Export.table.toDrive(
            means, description=to_drive, selectors=fields
        ).start()


def gee_soilgrids(band):
    """download and preprocess soilgrids from google earth engine

    https://git.wur.nl/isric/soilgrids/soilgrids.notebooks/-/blob/master/markdown/access_on_gee.md

    Args:
        band (string): one of ['bdod', 'cfvo', 'clay', 'sand', 'silt']

    Returns:
        image: ee.Image contains specific band
    """
    import ee

    image = ee.Image('projects/soilgrids-isric/' + band + '_mean')

    # calculate depth-weighted value
    RAW_NAMES = ['B5', 'B10', 'B15', 'B30', 'B40', 'B100']
    RAW_BANDS = [
        '0-5cm_mean',
        '5-15cm_mean',
        '15-30cm_mean',
        '30-60cm_mean',
        '60-100cm_mean',
        '100-200cm_mean',
    ]
    RAW_BANDS = [band + '_' + i for i in RAW_BANDS]
    RAW_DICT = {k: image.select(v) for k, v in zip(RAW_NAMES, RAW_BANDS)}
    image = image.expression(
        '5 * B5 + 10 * B10 + 15 * B15 + 30 * B30 + 40 * B40 + 100 * B100', RAW_DICT
    ).divide(200)
    image = image.reduceResolution(
        reducer=ee.Reducer.mean(), bestEffort=True, maxPixels=64
    )

    return image


def gee_wind(image, name='wind_10m'):
    wind_10m = image.expression(
        'sqrt(u**2 + v**2)',
        {
            'u': image.select('u_component_of_wind_10m'),
            'v': image.select('v_component_of_wind_10m'),
        },
    ).rename(name)
    time = image.get('system:time_start')
    return wind_10m.set('system:time_start', time)


def gee_group_by_month(images):
    # Group by month, and then reduce within groups by mean()
    # the result is an ImageCollection with one image for each
    # month.
    import ee

    months = ee.List.sequence(1, 12)
    by_month = ee.ImageCollection.fromImages(
        months.map(
            lambda m: images.filter(ee.Filter.calendarRange(m, m, 'month'))
            .mean()
            .set('month', m)
        )
    )
    return by_month


def gee_seasonality_index(images):
    avr = images.mean()
    SI = (
        gee_group_by_month(images)
        .map(lambda im: im.subtract(avr).abs())
        .sum()
        .divide(avr.multiply(12))
    )
    return SI



import os
import numpy as np
from osgeo import gdal
from collections.abc import Iterable
from geospace._const import WGS84, CREATION
from geospace.projection import read_srs, coord_trans


def block_write(datasets, map_fun, *args, **kwargs):
    import psutil

    if not isinstance(datasets, Iterable):
        datasets = [datasets]
    datasets = [ds_name(ds)[0] for ds in datasets]
    ds = datasets[0]
    n_x, n_y, band = ds.RasterXSize, ds.RasterYSize, ds.GetRasterBand(1)
    nodata = band.GetNoDataValue()
    ratio = int(
        n_y
        * n_x
        * ds.RasterCount
        * len(datasets)
        / (
            psutil.virtual_memory().available
            * 0.5
            / (2 + ds.ReadAsArray(0, 0, 1, 1).dtype.itemsize)
        )
        + 1
    )
    if ratio == 1:
        arrs = []
        for d in datasets:
            arr = d.ReadAsArray()
            mask = (arr == d.GetRasterBand(1).GetNoDataValue(),)
            arrs.append(np.ma.masked_array(arr, mask=mask, fill_value=nodata))
        arr = map_fun(arrs, *args, **kwargs)
        ds.WriteArray(arr)
        del arr, arrs, mask
    else:
        b_xsize, b_ysize = band.GetBlockSize()
        xsize = max(n_x // ratio + 1, b_xsize)
        ysize = max(n_y // ratio + 1, b_ysize)
        xoffs, yoffs = range(0, n_x, xsize), range(0, n_y, ysize)
        for xoff in xoffs:
            for yoff in yoffs:
                win_xsize = min(xsize, n_x - xoff)
                win_ysize = min(ysize, n_y - yoff)

                arrs = []
                for d in datasets:
                    arr = d.ReadAsArray(
                        xoff=xoff, yoff=yoff, xsize=win_xsize, ysize=win_ysize
                    )
                    mask = (arr == d.GetRasterBand(1).GetNoDataValue(),)
                    arrs.append(np.ma.masked_array(arr, mask=mask, fill_value=nodata))
                part_args = (
                    a[yoff : yoff + win_ysize, xoff : xoff + win_xsize] for a in args
                )
                part_kwargs = {
                    k: v[yoff : yoff + win_ysize, xoff : xoff + win_xsize]
                    for k, v in kwargs.items()
                }
                arr = map_fun(arrs, *part_args, **part_kwargs)
                ds.WriteArray(arr, xoff=xoff, yoff=yoff)
                del arr, arrs, mask, part_args, part_kwargs


def rep_file(cache_dir, filename):
    prefix, extension = os.path.splitext(os.path.basename(filename))
    file_path = os.path.join(cache_dir, prefix + extension)
    if os.path.exists(file_path):
        i = 1
        while True:
            file_path = os.path.join(cache_dir, prefix + '(' + str(i) + ')' + extension)
            if os.path.exists(file_path):
                i += 1
            else:
                break
    return file_path


def rep_name(rasters, sort_idxs=None):
    n_bands = [gdal.Open(ras).RasterCount for ras in rasters]
    t = np.cumsum(n_bands)
    s = np.roll(t, 1)
    s[0] = 0

    names = np.zeros(t[-1], dtype='object')
    for i, ras in enumerate(rasters):
        string = os.path.splitext(os.path.basename(ras))[0]
        names[s[i] : t[i]] = np.strings.add(
            string, np.strings.mod('%d', np.arange(n_bands[i]))
        )
        names[s[i]] = string

    if sort_idxs is None:
        return names, s, t

    relative_loc = np.zeros(t[-1], dtype=int)
    idxs = sort_idxs * np.power(10, int(np.ceil(np.log10(np.max(n_bands)))))
    for i, _ in enumerate(rasters):
        relative_loc[s[i] : t[i]] = idxs[i] + np.arange(n_bands[i])
    return names, s, t, np.argsort(relative_loc)


def geo2imagexy(ds, x, y):
    ds = ds_name(ds)[0]
    trans = ds.GetGeoTransform()
    a = np.array([[trans[1], trans[2]], [trans[4], trans[5]]])
    b = np.array([x - trans[0], y - trans[3]])
    col, row = np.linalg.solve(a, b) - 0.5
    return np.round(col).astype(int), np.round(row).astype(int)


def imagexy2geo(ds, row, col):
    """
    row, col to centroid lon, lat
    """
    ds = ds_name(ds)[0]
    trans = ds.GetGeoTransform()
    px = trans[0] + (col + 0.5) * trans[1] + (row + 0.5) * trans[2]
    py = trans[3] + (col + 0.5) * trans[4] + (row + 0.5) * trans[5]
    return px, py


def meshgrid(ds, geo_srs=WGS84):
    ds, _ = ds_name(ds)
    col, row = np.meshgrid(np.arange(ds.RasterXSize), np.arange(ds.RasterYSize))
    x, y = imagexy2geo(ds, row, col)

    if 'PROJCS' not in ds.GetProjection():
        return x, y

    trans = coord_trans(ds, geo_srs)
    lonlat = np.array(
        trans.TransformPoints(
            np.concatenate([x.reshape(-1, 1), y.reshape(-1, 1)], axis=1)
        )
    )
    lon = lonlat[:, 0].reshape(x.shape)
    lat = lonlat[:, 1].reshape(y.shape)
    return lon, lat


def context_file(ras, out_path):
    out_path = str(out_path)
    ext = os.path.splitext(os.path.basename(out_path))[1]
    if ext != '.tif':
        if (not os.path.isdir(out_path)) and ('/vsimem' not in out_path):
            os.makedirs(out_path)
        out_file = os.path.join(
            out_path, os.path.splitext(os.path.basename(ras))[0] + '.tif'
        )
    else:
        out_file = out_path

    return out_file


def ds_name(ds):
    if isinstance(ds, gdal.Dataset):
        ras = ds.GetDescription()
    else:
        ras = str(ds)
        ds = gdal.Open(ras)

    return ds, ras


def get_extent(layer):
    import json

    r = layer.GetSpatialRef()
    crs = json.loads(r.ExportToPROJJSON())
    axis = [
        crs['coordinate_system']['axis'][i]['direction'].upper()
        for i in range(r.GetAxesCount())
    ]
    order = ['EAST', 'WEST', 'NORTH', 'SOUTH']
    order_dict = dict(zip(order, range(4)))
    axis_code = np.array([order_dict[i] for i in axis])
    axis_sort = axis_code.argsort()
    data_map = [abs(i) - 1 for i in r.GetDataAxisToSRSAxisMapping()]
    extent = np.reshape(layer.GetExtent(), [-1, 2])[data_map]
    xy_extent = np.reshape(extent[axis_sort], [-1])
    return tuple(xy_extent)


def zeros_tif(out_file, x_size, y_size, n_band, data_type, trans, srs, nodata=2):
    ds = gdal.GetDriverByName('GTiff').Create(
        out_file, x_size, y_size, n_band, data_type, CREATION
    )

    # fill with 0 and set no data
    band = ds.GetRasterBand(1)
    band.Fill(0)
    band.SetNoDataValue(nodata)

    # set geotransform
    ds.SetGeoTransform(tuple(trans))

    # set SpatialReference
    ds.SetSpatialRef(read_srs(srs))

    return out_file



import os
import numpy as np
from pathlib import Path
from osgeo import gdal, ogr
from functools import partial
from geospace.projection import read_srs
from geospace.boundary import _enlarge_bound
from geospace.spatial_calc import area_per_row
from geospace._const import WGS84, CREATION, N_CPU
from geospace.shape import shp_projection, shp_filter
from geospace.utils import zeros_tif, block_write, context_file, ds_name, rep_name


# Now Read the large raster block by block, expecting to see no increase of
# memory as we loop through the blocks.
def _map_burn(arrs, burn_data):
    mask = np.broadcast_to(np.logical_not(burn_data.astype(bool)), arrs[0].shape)
    rect_data = np.ma.masked_where(mask, arrs[0])
    return rect_data.filled()


def _in_shape(rect_trans, enlarge, n_x, n_y, outLayer, rasterize_option):
    # set geotransform for enlarged tif
    poly_trans = list(rect_trans)
    poly_trans[1] = poly_trans[1] / enlarge
    poly_trans[5] = poly_trans[5] / enlarge

    # Rasterize
    poly_file = '/vsimem/_poly.tif'
    zeros_tif(
        poly_file,
        n_x * enlarge,
        n_y * enlarge,
        1,
        gdal.GDT_Byte,
        poly_trans,
        outLayer.GetSpatialRef(),
        nodata=2,
    )
    poly_ds = gdal.Open(poly_file, gdal.GA_Update)
    gdal.RasterizeLayer(
        poly_ds, [1], outLayer, burn_values=[1], options=rasterize_option
    )
    poly_data = poly_ds.ReadAsArray()

    # https://towardsdatascience.com/efficiently-splitting-an-image-into-tiles-in-python-using-numpy-d1bf0dd7b6f7
    if enlarge == 1:
        count = poly_data
    else:
        count = poly_data.reshape(n_y, enlarge, n_x, enlarge).swapaxes(1, 2)
        count = count.sum(axis=(-2, -1), dtype=np.uint16)
    return count


def _clip(
    ds,
    outLayer,
    out_file,
    ext='',
    enlarge=None,
    save_cache=False,
    reuse_cache=False,
    rasterize_option=['ALL_TOUCHED=TRUE'],
):
    # auto downscale the raster if resolution > 50 meter
    t = ds.GetGeoTransform()
    if enlarge is None:
        if 'PROJCS' in ds.GetProjection():
            enlarge = 1 if abs(t[1]) < 50 else 10
        else:
            enlarge = 1 if abs(t[1]) < 0.0005 else 10

    # get no data and Spatial Reference
    nodata = ds.GetRasterBand(1).GetNoDataValue()
    srs = outLayer.GetSpatialRef()

    # get shp extent in form of raster grids
    x_min, x_max, y_min, y_max = outLayer.GetExtent()
    bound, clip_range = _enlarge_bound(ds, x_min, y_min, x_max, y_max)
    n_x, n_y = int(clip_range[2]), int(clip_range[3])

    if out_file is not None:
        # clip with rectangle out_file, use Warp instead of Translate
        # to project longitude from (0, 360) to (-180, 180)
        option = gdal.WarpOptions(
            multithread=True,
            outputBounds=bound,
            srcSRS=srs,
            dstSRS=srs,
            creationOptions=CREATION,
            dstNodata=nodata,
            xRes=t[1],
            yRes=t[5],
            srcNodata=nodata,
            resampleAlg=gdal.GRA_NearestNeighbour,
        )
        ds_rect = gdal.Warp(out_file, ds, options=option)
        # mask value outside the shapefile
        is_in = _in_shape(
            ds_rect.GetGeoTransform(), 1, n_x, n_y, outLayer, rasterize_option
        )
        block_write(ds_rect, _map_burn, is_in)
        return ds_rect, is_in

    # get a rectangle containing the shapefile
    rect_trans = list(t)
    rect_trans[0], rect_trans[3] = bound[0], bound[3]
    # longitude from (0, 360) to (-180, 180)
    xoff = clip_range[0] if clip_range[0] > 0 else clip_range[0] + ds.RasterXSize
    xoff, yoff = int(xoff), int(clip_range[1])
    # shape cross the prime meridian
    if xoff + n_x > ds.RasterXSize:
        rect = np.concatenate(
            [
                ds.ReadAsArray(xoff, yoff, ds.RasterXSize - xoff, n_y),
                ds.ReadAsArray(0, yoff, xoff + n_x - ds.RasterXSize, n_y),
            ],
            axis=-1,
        )
    else:
        rect = ds.ReadAsArray(xoff, yoff, n_x, n_y)

    # paths of cached _burn.tif
    if save_cache:
        reuse_cache = True
        cache_dir = 'cache'
        if not os.path.exists(cache_dir):
            os.mkdir(cache_dir)
        burn_file = os.path.join(cache_dir, str(ext) + '_burn.tif')
    else:
        burn_file = os.path.join('/vsimem/', str(ext) + '_burn.tif')

    # calculate shapefile intersection area in each grid
    burn_ds = gdal.Open(burn_file)
    if (
        reuse_cache
        and (burn_ds is not None)
        and np.array_equal(rect_trans, burn_ds.GetGeoTransform())
        and srs.ExportToProj4() == burn_ds.GetSpatialRef().ExportToProj4()
        and n_x == burn_ds.RasterXSize
        and n_y == burn_ds.RasterYSize
    ):
        burn_data = burn_ds.ReadAsArray()
    else:
        # get counts of shape intersection for each grid
        count = _in_shape(rect_trans, enlarge, n_x, n_y, outLayer, rasterize_option)

        # calculate area for each grid
        row_area = area_per_row(rect_trans, srs.ExportToWkt(), n_y)
        burn_data = count * row_area.reshape([n_y, -1]) / enlarge / enlarge

        # save the burn_data into tif
        if reuse_cache:
            burn_ds = gdal.GetDriverByName('GTiff').Create(
                burn_file, n_x, n_y, 1, gdal.GDT_Float64, CREATION
            )
            burn_ds.SetGeoTransform(tuple(rect_trans))
            burn_ds.SetSpatialRef(srs)
            burn_ds.GetRasterBand(1).SetNoDataValue(0)
            burn_ds.WriteArray(burn_data)
            burn_ds = None

    return rect, burn_data


def extract(ras, shp, out_path=None, ras_srs=WGS84, nodata=None, **kwargs):
    ds, ras = ds_name(ras)
    if out_path is None:
        out_file = None
    else:
        out_file = context_file(ds.GetDescription(), out_path)
        if os.path.exists(out_file):
            return out_file

    # set projection
    inDataset = shp if isinstance(shp, gdal.Dataset) else ogr.Open(shp)
    inLayer = inDataset.GetLayer()
    if (
        read_srs([ds, ras_srs]).ExportToProj4()
        == inLayer.GetSpatialRef().ExportToProj4()
    ):
        outLayer = inLayer
    else:
        outDataSet = shp_projection(inLayer, False, out_srs=read_srs([ds, ras_srs]))
        outLayer = outDataSet.GetLayer()

    # set no data
    if ds.GetRasterBand(1).GetNoDataValue() is not None:
        nodata = ds.GetRasterBand(1).GetNoDataValue()
    elif nodata is not None:
        ds.GetRasterBand(1).SetNoDataValue(nodata)
    else:
        raise (ValueError('nodata must be initialized'))

    # clip with the whole shapefile
    rect, burn_data = _clip(ds, outLayer, out_file, **kwargs)
    if out_file is not None:
        return rect.GetDescription()

    # area weighted statistics
    not_in = np.broadcast_to(np.logical_not(burn_data.astype(bool)), rect.shape)
    mask = (rect == nodata) | not_in | (~np.isfinite(rect))
    arr = np.ma.masked_array(rect, mask)
    arr = arr[np.newaxis, :, :] if arr.ndim != 3 else arr

    return np.ma.average(
        arr.reshape(arr.shape[0], -1), weights=burn_data.ravel(), axis=1
    ).filled(np.nan)


def basin_average_worker(rasters, shp, is_unique, s, t, field, filter, **kwargs):
    if field == 'FID':
        filter_sql = f'{field} = {filter}'
    else:
        filter_sql = f"{field} = '{filter}'"
    ds_shp = shp_filter(shp, filter_sql, filter_shp=False)
    one_out = np.full(t[-1], np.nan)
    for i, ras in enumerate(rasters):
        kwargs['reuse_cache'] = False if is_unique[i] else True
        one_out[s[i] : t[i]] = extract(ras, ds_shp, ext=filter, **kwargs)
    return one_out


def basin_average(rasters, shp, field='STAID', filter=None, **kwargs):
    import tqdm
    import pandas as pd
    from multiprocessing import get_context

    if isinstance(rasters, (str, Path)):
        rasters = [str(rasters)]
    else:
        rasters = [str(ras) for ras in rasters]

    if filter is None:
        ds = ogr.Open(shp)
        layer = ds.GetLayer()
        filter = range(layer.GetFeatureCount())
        field = 'FID'
    if isinstance(filter, (str, int)):
        filter = [filter]

    arr = np.array([gdal.Open(ras).GetGeoTransform() for ras in rasters])
    sort_idxs = np.lexsort((arr[:, 0], arr[:, 3], arr[:, 1], arr[:, 5]))
    sort_rasters = np.array(rasters)[sort_idxs]
    _, idxs, counts = np.unique(
        arr[:, [1, 5]][sort_idxs], axis=0, return_index=True, return_counts=True
    )
    is_unique = np.full(len(rasters), False)
    is_unique[idxs[counts == 1]] = True

    sort_names, s, t, inverse = rep_name(sort_rasters, sort_idxs=sort_idxs)

    cpu_used = max(min(N_CPU, len(filter)), 1)
    if cpu_used == 1 or kwargs.pop('parallel', True) == False:
        output = list(
            tqdm.tqdm(
                (
                    partial(
                        basin_average_worker,
                        sort_rasters,
                        shp,
                        is_unique,
                        s,
                        t,
                        field,
                        **kwargs,
                    )(f)
                    for f in filter
                ),
                total=len(filter),
            )
        )
    else:
        with get_context('spawn').Pool(cpu_used) as p:
            output = list(
                tqdm.tqdm(
                    p.imap(
                        partial(
                            basin_average_worker,
                            sort_rasters,
                            shp,
                            is_unique,
                            s,
                            t,
                            field,
                            **kwargs,
                        ),
                        filter,
                    ),
                    total=len(filter),
                )
            )

    # use column to store array for performance
    # df = pd.DataFrame(output, columns=sort_names, index=filter)
    df = pd.DataFrame(np.vstack(output).T, columns=filter, index=sort_names)

    return df.iloc[inverse, :]



import os
from osgeo import gdal, ogr
from geospace.projection import read_srs


def polygonize(ds, shp_path):
    if os.path.exists(shp_path):
        return
    # mapping between gdal type and ogr field type
    type_mapping = {
        gdal.GDT_Byte: ogr.OFTInteger,
        gdal.GDT_UInt16: ogr.OFTInteger,
        gdal.GDT_Int16: ogr.OFTInteger,
        gdal.GDT_UInt32: ogr.OFTInteger,
        gdal.GDT_Int32: ogr.OFTInteger,
        gdal.GDT_Float32: ogr.OFTReal,
        gdal.GDT_Float64: ogr.OFTReal,
        gdal.GDT_CInt16: ogr.OFTInteger,
        gdal.GDT_CInt32: ogr.OFTInteger,
        gdal.GDT_CFloat32: ogr.OFTReal,
        gdal.GDT_CFloat64: ogr.OFTReal,
    }

    srcband = ds.GetRasterBand(1)
    dst_layername = 'Shape'
    drv = ogr.GetDriverByName('ESRI Shapefile')
    dst_ds = drv.CreateDataSource(shp_path)

    dst_layer = dst_ds.CreateLayer(dst_layername, srs=read_srs(ds))
    raster_field = ogr.FieldDefn('id', type_mapping[srcband.DataType])
    dst_layer.CreateField(raster_field)
    gdal.Polygonize(srcband, srcband, dst_layer, 0, [], callback=None)
    return shp_path



import os
import numpy as np
from osgeo import gdal
from pathlib import Path
from geospace.raster import mosaic
from collections.abc import Iterable
from geospace._const import CREATION, N_CPU
from geospace.utils import ds_name, context_file


def band_map(i, ras_multi, band_idx_multi, calc_arg, out_file):
    from osgeo_utils.gdal_calc import Calc

    ds_ras = gdal.Open(ras_multi[0])
    band_name = ds_ras.GetRasterBand(int(band_idx_multi[0])).GetDescription()
    tem_name = band_name if band_name else f'_temp_{Path(out_file).stem}_{str(i)}'
    tem_file = f'{Path(out_file).parent}/{tem_name}.tif'

    if os.path.exists(tem_file):
        return tem_file

    ras_args = {chr(i + 65): ras for i, ras in enumerate(ras_multi)}
    band_args = {
        chr(i + 65) + '_band': int(band_idx)
        for i, band_idx in enumerate(band_idx_multi)
    }
    input_args = {**ras_args, **band_args}

    Calc(calc_arg, tem_file, creation_options=CREATION, quiet=True, **input_args)

    return tem_file


def check_iter(ds_multi, calc_args, band_idxs):
    iter_ds_multi = isinstance(ds_multi, Iterable) and not isinstance(ds_multi, str)
    iter_calc_args = isinstance(calc_args, Iterable) and not isinstance(calc_args, str)
    if band_idxs is not None:
        if iter_ds_multi:
            iter_band_idxs = isinstance(band_idxs[0], Iterable)
        else:
            iter_band_idxs = isinstance(band_idxs, Iterable)
    else:
        iter_band_idxs = False

    return iter_ds_multi, iter_calc_args, iter_band_idxs


def broadcast_args(ds_multi, calc_args, band_idxs):
    iter_ds_multi, iter_calc_args, iter_band_idxs = check_iter(
        ds_multi, calc_args, band_idxs
    )

    if iter_ds_multi:
        ds = ds_multi[0]
    else:
        ds = ds_multi

    ds, ras = ds_name(ds)
    if band_idxs is not None:
        if iter_band_idxs and iter_calc_args:
            if len(band_idxs) != len(calc_args):
                raise Exception('length of band list not equal to that of calc args')
        elif iter_band_idxs:
            calc_args = [calc_args] * len(band_idxs)
        elif iter_calc_args:
            band_idxs = [band_idxs] * len(calc_args)
        else:
            calc_args = [calc_args]
            band_idxs = [band_idxs]
    else:
        n_band = ds.RasterCount
        if iter_calc_args:
            if len(calc_args) != n_band:
                raise Exception('calc args length not equal to band counts')
        else:
            calc_args = [calc_args] * n_band

        if iter_ds_multi:
            band_idxs = np.repeat(
                np.arange(1, n_band + 1, dtype=int), len(ds_multi)
            ).reshape(-1, len(ds_multi))
        else:
            band_idxs = np.arange(1, n_band + 1, dtype=int)

    if not iter_ds_multi:
        ds_multi = [ras]
        band_idxs = np.array(band_idxs).reshape(len(band_idxs), 1)

    return ds_multi, calc_args, band_idxs


def map_calc(ds_multi, calc_args, out_path, band_idxs=None, parallel=True):
    import tqdm
    from multiprocessing import get_context

    if isinstance(ds_multi, Iterable) and not isinstance(ds_multi, str):
        ds_multi = [str(ds) for ds in ds_multi]
        ds = ds_multi[0]
    else:
        ds_multi = str(ds_multi)
        ds = ds_multi

    ds, ras = ds_name(ds)
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    ds_multi, calc_args, band_idxs = broadcast_args(ds_multi, calc_args, band_idxs)

    n = len(calc_args)
    args = zip(
        np.arange(1, n + 1, dtype=int),
        [ds_multi] * n,
        band_idxs,
        calc_args,
        [out_file] * n,
    )

    if parallel and (n > 1):
        with get_context('spawn').Pool(min(N_CPU, n)) as p:
            tem_files = p.starmap(band_map, tqdm.tqdm(args, total=n))
    else:
        tem_files = []
        for arg in args:
            tem_files.append(band_map(*arg))

    if len(tem_files) == 1:
        os.rename(tem_files[0], out_file)
    else:
        mosaic(tem_files, out_file, separate=True)
        [os.remove(f) for f in tem_files]

    return out_file
